import os
import subprocess
from collections import defaultdict
from shutil import which

import boto3
import pytest
import yaml
from re import sub
from yaml.scanner import ScannerError

_serverless_yml_dict = None


def _get_property(properties, property_names):
    return {name: properties[name] for name in property_names if name in properties}


def _handle_dynamodb_table(resources):
    from moto import mock_dynamodb

    dynamodb = mock_dynamodb()

    def before():
        dynamodb.start()

        for resource_definition in resources:
            boto3.resource("dynamodb").create_table(
                **_get_property(
                    resource_definition["Properties"],
                    (
                        "TableName",
                        "AttributeDefinitions",
                        "KeySchema",
                        "LocalSecondaryIndexes",
                        "GlobalSecondaryIndexes",
                        "BillingMode",
                        "ProvisionedThroughput",
                        "StreamSpecification",
                        "SSESpecification",
                        "Tags",
                    ),
                )
            )

    def after():
        for resource_definition in resources:
            boto3.client("dynamodb").delete_table(
                TableName=resource_definition["Properties"]["TableName"]
            )

        dynamodb.stop()

    return before, after


def _handle_sqs_queue(resources):
    from moto import mock_sqs

    sqs = mock_sqs()

    def before():
        sqs.start()

        for resource_definition in resources:
            boto3.resource("sqs").create_queue(
                QueueName=resource_definition["Properties"]["QueueName"]
            )

    def after():
        sqs_client = boto3.client("sqs")
        for resource_definition in resources:
            sqs_client.delete_queue(
                QueueUrl=sqs_client.get_queue_url(
                    QueueName=resource_definition["Properties"]["QueueName"]
                )["QueueUrl"]
            )

        sqs.stop()

    return before, after


def _handle_s3_bucket(resources):
    from moto import mock_s3

    s3 = mock_s3()

    def before():
        s3.start()

        for resource_definition in resources:
            if resource_definition["Properties"].get("BucketName"):
                boto3.resource("s3").create_bucket(
                    Bucket=resource_definition["Properties"]["BucketName"],
                    **_get_property(
                        resource_definition["Properties"],
                        (
                            "CreateBucketConfiguration",
                            "ACL",
                            "GrantFullControl",
                            "GrantRead",
                            "GrantReadACP",
                            "GrantWrite",
                            "GrantWriteACP",
                            "ObjectLockEnabledForBucket",
                        ),
                    ),
                )

    def after():
        s3_resource = boto3.resource("s3")

        for resource_definition in resources:
            if resource_definition["Properties"].get("BucketName"):
                bucket = s3_resource.Bucket(
                    resource_definition["Properties"]["BucketName"]
                )
                bucket.object_versions.delete()

    return before, after


def _handle_sns_topic(resources):
    from moto import mock_sns

    sns = mock_sns()

    def before():
        sns.start()

        for resource_definition in resources:
            if resource_definition["Properties"].get("TopicName"):
                boto3.resource("sns").create_topic(
                    Name=resource_definition["Properties"]["TopicName"]
                )

    def after():
        sns_client = boto3.client("sns")

        topic_arns = {
            arn["TopicArn"].split(":")[-1]: arn["TopicArn"]
            for arn in sns_client.list_topics()["Topics"]
        }

        for resource_definition in resources:
            if resource_definition["Properties"].get("TopicName"):
                sns_client.delete_topic(
                    TopicArn=topic_arns[resource_definition["Properties"]["TopicName"]]
                )

        sns.stop()

    return before, after


def _handle_kms_key(resources):
    from moto import mock_kms

    kms = mock_kms()

    def before():
        kms.start()

        for resource_definition in resources:
            params = _get_property(
                resource_definition["Properties"], ("Description", "KeyUsage", "Tags")
            )

            if "KeyPolicy" in resource_definition["Properties"]:
                params["Policy"] = resource_definition["Properties"]["KeyPolicy"]

            if "KeySpec" in resource_definition["Properties"]:
                params["CustomerMasterKeySpec"] = resource_definition["Properties"][
                    "KeySpec"
                ]

            boto3.client("kms").create_key(**params)

    def after():
        kms.stop()

    return before, after


SUPPORTED_RESOURCES = {
    "AWS::DynamoDB::Table": _handle_dynamodb_table,
    "AWS::SQS::Queue": _handle_sqs_queue,
    "AWS::S3::Bucket": _handle_s3_bucket,
    "AWS::SNS::Topic": _handle_sns_topic,
    "AWS::KMS::Key": _handle_kms_key,
}


@pytest.fixture()
def serverless():
    global _serverless_yml_dict
    if not _serverless_yml_dict:
        _serverless_yml_dict = _load_file()

    for k, v in _serverless_yml_dict["provider"].get("environment", {}).items():
        os.environ[k] = v

    actions_before = []
    actions_after = []

    resources = defaultdict(list)
    for resource_name, definition in (
        _serverless_yml_dict.get("resources", {}).get("Resources", {}).items()
    ):
        resources[definition["Type"]].append(definition)

    for resource_name, resource_function in SUPPORTED_RESOURCES.items():
        if resources.get(resource_name):
            resource = resource_function(resources[resource_name])
            actions_before.append(resource[0])
            actions_after.append(resource[1])

    for action in actions_before:
        action()

    yield

    for action in actions_after:
        action()


def _load_file():
    serverless_path = os.environ.get("SERVERLESS_FILE_PATH", "serverless.yml")
    serverless_command = os.environ.get("SERVERLESS_COMMAND", "sls")

    is_serverless = os.path.isfile(serverless_path)
    if not is_serverless:
        raise Exception("No serverless.yml file found!")

    if serverless_command not in ("sls", "serverless"):
        raise Exception("Command %s not releated to serverless" % serverless_command)

    if not which(serverless_command):
        raise Exception("No %s executable found!" % serverless_command)

    env = os.environ.copy()
    env["SLS_DEPRECATION_DISABLE"] = "*"
    env["SLS_WARNING_DISABLE"] = "*"
    result = subprocess.run([serverless_command, "print", "--config", serverless_path], stdout=subprocess.PIPE, env=env)
    serverless_content = sub(r'Serverless: [^\n]*\n', "", result.stdout.decode("utf-8"))

    try:
        return yaml.safe_load(serverless_content)
    except ScannerError as e:
        pytest.fail(
            f"serverless.yml is wrongly formatted, pytest-serverless is unable to load it: {e}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['pytest_serverless']
install_requires = \
['boto3>=1.22,<2.0', 'moto>=3.1,<4.0', 'pyyaml>=6.0,<7.0']

entry_points = \
{'pytest11': ['serverless = pytest_serverless']}

setup_kwargs = {
    'name': 'pytest-serverless',
    'version': '1.0.0',
    'description': 'Automatically mocks resources from serverless.yml in pytest using moto.',
    'long_description': 'pytest-serverless\n---\nAutomatically mocks resources defined in serverless.yml file using [moto](https://github.com/spulec/moto) and uses them in [pytest](https://github.com/pytest-dev/pytest).\n\nThis way you can focus on writing tests rather than defining enormous list of fixtures.\n\n| master | PyPI | Python | pytest | Licence |\n| --- | --- | --- | --- | --- |\n| ![Master](https://github.com/whisller/pytest-serverless/workflows/Master/badge.svg) | [![PyPI](https://img.shields.io/pypi/v/pytest-serverless.svg)](https://pypi.org/project/pytest-serverless/) | ![](https://img.shields.io/pypi/pyversions/pytest-serverless.svg) | `6.2` | ![](https://img.shields.io/pypi/l/pytest-serverless.svg) |\n\n## Pre installation requirements\n- `serverless` installed\n- `pytest` installed\n\n## Installation\n```sh\npip install pytest-serverless\n```\n\n## Usage\nAssuming your `serverless.yml` file looks like:\n```yaml\nservice: my-microservice\nresources:\n Resources:\n   TableA:\n     Type: \'AWS::DynamoDB::Table\'\n     DeletionPolicy: Delete\n     Properties:\n       TableName: ${self:service}.my-table\n       AttributeDefinitions:\n         - AttributeName: id\n           AttributeType: S\n         - AttributeName: company_id\n           AttributeType: S\n       KeySchema:\n         - AttributeName: id\n           KeyType: HASH\n       GlobalSecondaryIndexes:\n         - IndexName: company_id\n           KeySchema:\n             - AttributeName: company_id\n               KeyType: HASH\n           Projection:\n             ProjectionType: ALL\n           ProvisionedThroughput:\n             ReadCapacityUnits: 10\n             WriteCapacityUnits: 30\n       ProvisionedThroughput:\n         ReadCapacityUnits: 10\n         WriteCapacityUnits: 30\n```\n\nJust mark your test with `@pytest.mark.usefixtures("serverless")` and `pytest-serverless` will automatically create `my-microservice.my-table` dynamodb table.\n```python\nimport boto3\nimport pytest\n\n\n@pytest.mark.usefixtures("serverless")\ndef test():\n    table = boto3.resource("dynamodb").Table("my-microservice.my-table")\n    count_of_items = len(table.scan()["Items"])\n    assert count_of_items == 0\n```\n\nYou can use a custom serverless file path setting the envionmnet variable `SERVERLESS_FILE_PATH`.\n\n```shell\n$ export SERVERLESS_FILE_PATH=/path/to/serverless.yml\n```\n\nYou can use choose both `sls` or `serverless` command to run, settings the environment variable `SERVERLESS_COMMAND`. It will only accpets `sls` or `serverless` values.\n\n```shell\n$ export SERVERLESS_COMMAND=sls\n```\n\n## Supported resources\n### AWS::DynamoDB::Table\n### AWS::SQS::Queue\n### AWS::SNS::Topic\n### AWS::S3::Bucket\n### AWS::KMS::Key\n',
    'author': 'Daniel Ancuta',
    'author_email': 'whisller@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/whisller/pytest-serverless',
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
